# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lightworks import State


def convert_fock_data_to_state(raw_state: str) -> State:
    """
    Takes raw fock state data and converts this to a Lightworks state.
    """
    try:
        if raw_state[0] == "(":
            raw_state = raw_state[1:-1]
        mode_data = raw_state.split(", ")
        if len(mode_data) == 1:
            mode_data[0] = mode_data[0].replace(",", "")
        state = State([int(s) for s in mode_data if s])
    except Exception as e:
        raise ValueError(
            "Unable to convert the provided State to a Lightworks state."
        ) from e
    return state



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import os
from random import randint

import pytest
import requests

from lightworks_remote import AuthenticationError, token
from lightworks_remote.remote import BatchJob, Job


class TestJob:
    """
    Unit tests for Job object, which handles submission of payload and
    communication with the system.
    """

    def setup_method(self):
        """
        Creates an example payload which can be used for testing.
        """
        self.test_payload = {
            "qpu": "Test",
            "input": [randint(1, 10) for _ in range(4)],
            "unitary": [[randint(1, 10) for _ in range(4)] for _ in range(4)],
            "post_selection": None,
        }

    def test_creation(self):
        """
        Checks a job can be created.
        """
        Job()

    def test_creation_payload(self):
        """
        Checks a job can be created with a provided test payload.
        """
        Job(self.test_payload)

    def test_payload_assignement(self):
        """
        Checks payload returned by property is the assigned payload on job
        creation.
        """
        j = Job(self.test_payload)
        assert j.payload == self.test_payload

    def test_job_id_before_submission(self):
        """
        Checks job id returns None if this is requested before submission.
        """
        j = Job(self.test_payload)
        assert j.job_id is None

    def test_job_submission(self):
        """
        Checks that the submission code is working up to the point at which it
        fails because of a connection or authentication error.
        """
        j = Job(self.test_payload)
        token.set("1234")
        with contextlib.suppress(requests.ConnectionError, AuthenticationError):
            j._submit()

    def test_job_save(self):
        """
        Checks that a job can be saved to a file when values are null.
        """
        j = Job()
        j.save_to_file("_test_data")
        assert os.path.isfile("_test_data.json")
        os.remove("_test_data.json")

    def test_job_save_overwrite(self):
        """
        Checks that a job is protected from being overwritten unless this is
        allowed.
        """
        j = Job()
        j.save_to_file("_test_data")
        assert os.path.isfile("_test_data.json")
        # Check with overwrite off
        with pytest.raises(ValueError):
            j.save_to_file("_test_data")
        # And then allow this
        j.save_to_file("_test_data", allow_overwrite=True)
        os.remove("_test_data.json")

    def test_job_save_with_data(self):
        """
        Checks that a job can be saved to a file when there is a job id and
        payload.
        """
        j = Job(self.test_payload)
        j._job_id = randint(1, 100)
        j.save_to_file("_test_data")
        assert os.path.isfile("_test_data.json")
        os.remove("_test_data.json")

    def test_job_load(self):
        """
        Checks that a job can be loaded from a file and the data is preserved.
        """
        j = Job()
        j.save_to_file("_test_data")
        new_j = Job()
        new_j.load_from_file("_test_data")
        assert new_j.job_id == j.job_id
        assert new_j.payload == j.payload
        os.remove("_test_data.json")

    def test_job_load_with_data(self):
        """
        Checks that a job can be loaded from a file and the data is preserved
        when a payload and job id are assigned.
        """
        j = Job(self.test_payload)
        j._job_id = randint(1, 100)
        j.save_to_file("_test_data")
        new_j = Job()
        new_j.load_from_file("_test_data")
        assert new_j.job_id == j.job_id
        assert new_j.payload.payload == j.payload
        os.remove("_test_data.json")

    def teardown_class(self):
        """
        Deletes _test_data in case this is missed.
        """
        with contextlib.suppress(FileNotFoundError):
            os.remove("_test_data.json")


class TestBatchJob:
    """
    Unit tests for BatchJob object.
    """

    def setup_method(self):
        """
        Creates an example payload and job which can be used for testing.
        """
        self.test_payload = {
            "qpu": "Test",
            "input": [randint(1, 10) for _ in range(4)],
            "unitary": [[randint(1, 10) for _ in range(4)] for _ in range(4)],
            "post_selection": None,
        }
        self.test_job = Job(self.test_payload)

    def test_batch_job_creation(self):
        """
        Checks BatchJob can be created.
        """
        BatchJob()

    def test_batch_job_addition(self):
        """
        Checks that a job can be added to a BatchJob
        """
        BatchJob()._add(self.test_job, "test")

    def test_job_iteration(self):
        """
        Checks that batch job can be iterated through.
        """
        job = BatchJob()
        job._add(self.test_job, "test1")
        job._add(self.test_job, "test2")
        for i, name in enumerate(job):
            assert name == f"test{i + 1}"

    def test_batch_job_jobs_property(self):
        """
        Adds multiple jobs to a BatchJob and confirms the names and job data
        is correctly stored.
        """
        job = BatchJob()
        job._add(self.test_job, "test1")
        job._add(self.test_job, "test2")
        assert len(job.jobs) == 2
        assert "test1" in job
        assert "test2" in job
        assert job["test1"] is self.test_job

    def test_batch_job_save(self):
        """
        Checks that a BatchJob can be saved from a file.
        """
        job = BatchJob()
        job._add(self.test_job, "test1")
        job._add(self.test_job, "test2")
        job.save_to_file("_test_batch_data")
        assert os.path.isfile("_test_batch_data.json")
        os.remove("_test_batch_data.json")

    def test_batch_job_save_overwrite(self):
        """
        Checks that a job is protected from being overwritten unless this is
        allowed.
        """
        job = BatchJob()
        job._add(self.test_job, "test1")
        job._add(self.test_job, "test2")
        job.save_to_file("_test_data")
        assert os.path.isfile("_test_data.json")
        # Check with overwrite off
        with pytest.raises(ValueError):
            job.save_to_file("_test_data")
        # And then allow this
        job.save_to_file("_test_data", allow_overwrite=True)
        os.remove("_test_data.json")

    def test_batch_job_load(self):
        """
        Checks that a BatchJob can be loaded from a file.
        """
        job = BatchJob()
        job._add(self.test_job, "test1")
        job._add(self.test_job, "test2")
        job.save_to_file("_test_batch_data")
        new_job = BatchJob()
        new_job.load_from_file("_test_batch_data")
        assert new_job.keys() == job.keys()
        assert new_job["test1"].payload["input"] == self.test_payload["input"]
        os.remove("_test_batch_data.json")

    def teardown_class(self):
        """
        Deletes _test_batch_data in case this is missed.
        """
        with contextlib.suppress(FileNotFoundError):
            os.remove("_test_batch_data.json")



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lightworks as lw
import pytest

from lightworks_remote import QPU
from lightworks_remote.payloads import Payload


class TestQPU:
    """
    Unit tests for QPU object.
    """

    def test_qpu_creation(self):
        """
        Checks that a QPU can be successfully created.
        """
        QPU("test")

    @pytest.mark.parametrize("value", [True, 1, ["Test"]])
    def test_invalid_qpu_name(self, value):
        """
        Checks that a QPU name must be a string.
        """
        with pytest.raises(TypeError):
            QPU(value)

    @pytest.mark.parametrize("value", [True, 1, ["Test"]])
    def test_invalid_qpu_name_property(self, value):
        """
        Checks that the QPU name property cannot be overwritten by a non-string
        value.
        """
        q = QPU("test")
        with pytest.raises(TypeError):
            q.name = value

    def test_qpu_name_in_str(self):
        """
        Checks that QPU name in return value when str used on the QPU.
        """
        q = QPU("test")
        assert "test" in str(q)

    def test_qpu_name_in_repr(self):
        """
        Checks that QPU name in return value when repr used on the QPU.
        """
        q = QPU("test")
        assert "test" in repr(q)

    def test_payload_type(self):
        """
        Checks that created payload is always of the correct type.
        """
        task = lw.Sampler(lw.PhotonicCircuit(4), lw.State([1, 0, 1, 0]), 1000)
        qpu = QPU("test")
        assert isinstance(
            qpu._build_payload_from_task(task, False, "test"), Payload
        )

    def test_job_name(self):
        """
        Checks that job name is correctly replicated in the Payload.
        """
        task = lw.Sampler(lw.PhotonicCircuit(4), lw.State([1, 0, 1, 0]), 1000)
        qpu = QPU("test")
        payload = qpu._build_payload_from_task(task, False, "test_1")
        assert payload.job_name == "test_1"

    def test_direct_encoding(self):
        """
        Checks that direct encoding is correctly replicated in the Payload and
        that a circuit spec is also introduced.
        """
        task = lw.Sampler(
            lw.Unitary(lw.random_unitary(4)), lw.State([1, 0, 1, 0]), 1000
        )
        qpu = QPU("test")
        payload = qpu._build_payload_from_task(task, True, "test_1")
        assert payload.direct_implementation
        assert payload.circuit_spec

    def test_job_options(self):
        """
        Checks that options are correctly replicated in the payload job_data.
        """
        task = lw.Sampler(lw.PhotonicCircuit(4), lw.State([1, 0, 1, 0]), 1000)
        qpu = QPU("test")
        options = {"1": "Test", "2": 3}
        payload = qpu._build_payload_from_task(task, False, "test_1", options)
        assert "1" in payload.job_data
        assert "2" in payload.job_data
        assert payload.job_data["1"] == "Test"
        assert payload.job_data["2"] == 3



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains utilities for assisting with saving and loading data from files.
"""

import json
import platform
from pathlib import Path
from typing import Any

from lightworks_remote.payloads import Payload

from .serialization import serialize


def write_data_to_json(
    data: Payload | dict[str, Any], filename: str | Path
) -> None:
    """
    Writes the provided data to a json file.
    """
    data = serialize(data)
    full_path = validate_json_filename(filename)
    # Create directory if it doesn't exist
    full_path.parent.mkdir(parents=True, exist_ok=True)

    with open(full_path, "w", encoding="locale") as f:
        json.dump(data, f, indent=4)


def get_data_from_json(filename: str | Path) -> dict[str, Any]:
    """
    Gets data from a json file.
    """
    full_path = validate_json_filename(filename)
    if not saved_job_exists(full_path):
        raise FileNotFoundError("File with provided name does not exist.")
    with open(full_path, encoding="locale") as f:
        return json.load(f)


def validate_json_filename(filename: str | Path) -> Path:
    """
    Checks that a provided file name is a string and adds .json at end if this
    is not included
    """
    if not isinstance(filename, str | Path):
        raise TypeError("Provided filename should be a string or Path.")
    return Path(filename).with_suffix(".json")


def saved_job_exists(filename: str | Path) -> bool:
    """
    Checks if a saved job with the provided filename exists.
    """
    full_path = validate_json_filename(filename)
    return full_path.exists()


def get_persistent_file_data_location() -> Path:
    """
    Gets OS-dependent location that any persistent data should be saved to.
    """
    if platform.system() == "Windows":
        path = Path.home() / ".aegiq/lightworks/"
    elif platform.system() == "Linux":
        path = Path.home() / ".local/share/Aegiq/lightworks/"
    elif platform.system() == "Darwin":
        path = Path.home() / "Library/Application Support/Aegiq/lightworks/"
    else:
        msg = (
            "Certificate save location not yet implemented for current "
            f"platform ({platform.system()})."
        )
        raise NotImplementedError(msg)
    path.mkdir(parents=True, exist_ok=True)
    return path



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import string
from random import choice

import pytest

from lightworks_remote import TokenError, token

options = list(string.ascii_letters) + [str(i) for i in range(9)]
TOKEN = "".join([choice(options) for _ in range(1000)])
TOKEN_SAVE_NAME = "pytest_test_token"  # noqa: S105


class TestToken:
    """
    Checks the correct functioning of the token module.
    """

    def setup_method(self):
        """Perform setup before each test in which the token value is reset."""
        token.set(TOKEN)

    def test_set(self):
        """Confirms token can be set to a provided value."""
        assert token._Token__token == TOKEN

    @pytest.mark.parametrize("value", [None, 1234, ["test"]])
    def test_invalid_set(self, value):
        """
        Checks TypeError raised if an invalid type is set for the token.
        """
        with pytest.raises(TypeError):
            token.set(value)

    def test_get(self):
        """Confirms token value can be retrieved using get."""
        assert token.get() == TOKEN

    def test_str(self):
        """Checks that str returns token value."""
        assert str(token) == TOKEN

    def test_repr(self):
        """Checks token value is in repr return."""
        assert TOKEN in repr(token)

    def test_save(self):
        """Checks that token is able to be saved successfully."""
        token.save(TOKEN_SAVE_NAME, overwrite=True)

    def test_save_no_overwrite(self):
        """
        Confirms an error is raised if a token is saved with the same name twice
        with the overwrite option being enabled.
        """
        token.save(TOKEN_SAVE_NAME, overwrite=True)
        with pytest.raises(TokenError):
            token.save(TOKEN_SAVE_NAME, overwrite=False)

    def test_save_default_no_overwrite(self):
        """
        Confirms that default value for the overwrite option if False by
        checking an exception is raised if token is saved with same name twice.
        """
        token.save(TOKEN_SAVE_NAME, overwrite=True)
        with pytest.raises(TokenError):
            token.save(TOKEN_SAVE_NAME)

    def test_load(self):
        """
        Checks that token is able to be overwritten and then retrieved using the
        load method.
        """
        token.save(TOKEN_SAVE_NAME, overwrite=True)
        token.set("None")
        token.load(TOKEN_SAVE_NAME)
        assert token.get() == TOKEN

    def teardown_class(self):
        """Perform tear down by deleting saved token."""
        with contextlib.suppress(TokenError):
            token.delete_saved(TOKEN_SAVE_NAME)



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from pathlib import Path
from typing import Any

from lightworks_remote.results import QPUSamplingResult
from lightworks_remote.token import token
from lightworks_remote.utils import (
    HTTPError,
    JobError,
    JobNotFoundError,
    JobSubmissionError,
    ResultsError,
    convert_fock_data_to_state,
    get_persistent_file_data_location,
)
from lightworks_remote.utils.configuration import URL

from .artemis_job_details import (
    ARTEMIS_JOB_DETAILS_MAP,
    ArtemisJobDetails,
    get_no_results_reason,
)
from .artemis_qpu_details import ARTEMIS_QPU_DETAILS_MAP, ArtemisQPU
from .interface import HTTPInterface
from .status import Status


class ArtemisHTTPInterface(HTTPInterface):
    """
    Handles all communication with Artemis cloud interface.
    """

    def __init__(self) -> None:
        certificate = self._get_certificate()
        super().__init__(URL, certificate)

    def submit_job(self, payload: dict[str, Any]) -> int:
        """
        Submits the provided payload dictionary to the configured url and
        returns the job id if accepted.

        Args:

            payload (dict) : Serializable dictionary for submission to the url.

        Returns:

            int : The created job id.

        """
        response = self._post(
            "jobs", headers={"x-access-token": token.get()}, json=payload
        )

        if response.status_code == 202:
            return response.headers["X-Job-Id"]
        if response.status_code == 400:
            r = response.json()
            error_msg = "The following fields failed validation: \n"
            for e in r["errors"]:
                error_msg += f"\t{e['name']} : {e['reason']}. \n"
            raise JobSubmissionError(error_msg)
        raise HTTPError(
            "Something unexpected went wrong while attempting to submit "
            "the job.",
            response=response,
        )

    def get_job_status(self, job_id: int | str) -> Status:
        """
        Returns status of job with the requested id.
        """
        return self.get_job_details(job_id).job_status

    def get_job_queue_position(self, job_id: int | str) -> int | None:
        """
        Returns queue position of the requested job if it has been scheduled,
        otherwise will return None.
        """
        details = self.get_job_details(job_id)
        return details.queue_position

    def get_job_details(self, job_id: int | str) -> ArtemisJobDetails:
        """Gets full set of details for the job with the requested id."""
        response = self._get(
            f"jobs/{job_id}", headers={"x-access-token": token.get()}
        )

        if response.status_code == 200:
            details = ArtemisJobDetails(
                **{
                    ARTEMIS_JOB_DETAILS_MAP[k]: v
                    for k, v in response.json().items()
                }
            )
            # Remove queue position if job not in queue
            if details.job_status != Status.SCHEDULED:
                details.queue_position = None
            return details
        if response.status_code == 404:
            raise JobNotFoundError("Job with requested ID not found.")
        raise HTTPError(
            "Something unexpected went wrong while attempting to retrieve "
            "job details.",
            response=response,
        )

    def check_for_compilation_errors(
        self, job_id: int | str
    ) -> tuple[bool, list[tuple[int, str]]]:
        """
        Checks if any errors may have occurred during the compilation process.
        Note: This doesn't check if compilation has actually taken place so
        should only be called after the compilation.
        """
        details = self.get_job_details(job_id)
        if details.job_status != Status.FAILED:
            return (False, [])
        for log in details.logs:
            if "CompilationErrors" in log["meta"]:
                errors = [
                    (e["Code"], e["Message"])
                    for e in log["meta"]["CompilationErrors"]
                ]
                return (True, errors)
        return (False, [])

    def cancel_job(self, job_id: int | str) -> None:
        """
        Cancels job with matching id, raising an error if a matching job is not
        found.
        """
        # First check if job exists and if it can be cancelled.
        details = self.get_job_details(job_id)
        if details.job_status == Status.CANCELLED:
            msg = (
                f"Job with requested ID ({job_id}) has already been cancelled."
            )
            raise JobError(msg)
        if not details.job_status.is_cancellable:
            msg = (
                f"Job with requested ID ({job_id}) is not in a cancellable "
                "state."
            )
            raise JobError(msg)
        # If passes these checks then try to cancel
        response = self._put(
            f"jobs/{job_id}/cancel", headers={"x-access-token": token.get()}
        )
        if response.status_code == 204:
            return
        raise HTTPError(
            "Something unexpected went wrong while attempting to cancel "
            "the job.",
            response=response,
        )

    def get_results(
        self, job_id: int | str, max_retry: int = 10
    ) -> QPUSamplingResult:
        """
        Retrieves the results for the job with the supplied id.
        """
        # First check job status shows job is complete
        details = self.get_job_details(job_id)
        if not details.job_status.is_complete:
            msg = f"Job with requested ID ({job_id}) is not yet complete."
            raise ResultsError(msg)
        if not details.job_status.has_results:
            reason = get_no_results_reason(details)
            msg = (
                f"Job with requested ID ({job_id}) did not generate any "
                f"results {reason}"
            )
            raise ResultsError(msg)
        # Then attempt to get results
        response = self._get(
            f"jobs/{job_id}/results", headers={"x-access-token": token.get()}
        )
        if response.status_code == 200:
            results = response.json()
            processed_result = {
                key: val
                if key != "results" or val is None
                else {
                    convert_fock_data_to_state(state): counts
                    for state, counts in val.items()
                }
                for key, val in results.items()
            }
            return QPUSamplingResult(processed_result)
        if response.status_code == 400:
            reason = get_no_results_reason(details)
            msg = (
                f"Job with requested ID ({job_id}) did not generate any "
                f"results {reason}"
            )
            raise ResultsError(msg)
        # Sometimes 404 is returned when results are requested too quickly, in
        # this case sleep and retry
        if response.status_code == 404:
            time.sleep(0.5)
            if max_retry > 0:
                return self.get_results(job_id, max_retry=max_retry - 1)
            msg = (
                f"Unable to retrieve results for Job with ID {job_id}, these "
                "may not be available yet."
            )
            raise ResultsError(msg)
        raise HTTPError(
            "Something unexpected went wrong while attempting to get results "
            "for the job.",
            response=response,
        )

    def list_all_job_ids(self) -> list[int]:
        """
        List all jobs submitted to the system.
        """
        response = self._get(
            "jobs",
            headers={"x-access-token": token.get()},
            params={"pageSize": 2**31 - 1},
        )
        return sorted(
            [data["id"] for data in response.json()["items"]], reverse=True
        )

    def list_scheduled_job_ids(self) -> list[int]:
        """
        Lists all jobs awaiting execution on the system.
        """
        response = self._get(
            "jobs",
            headers={"x-access-token": token.get()},
            params={
                "pageSize": 2**31 - 1,
                "statuses": ["Accepted", "Scheduled"],
            },
        )
        return sorted([data["id"] for data in response.json()["items"]])

    def list_all_job_details(self) -> list[dict[str, Any]]:
        """
        Returns details of all jobs submitted to system.
        """
        response = self._get(
            "jobs",
            headers={"x-access-token": token.get()},
            params={"pageSize": 2**31 - 1},
        )
        return response.json()["items"]

    def get_all_qpus(self) -> dict[str, dict[str, Any]]:
        """
        Returns details of all QPUs, such as names, ids and status.
        """
        response = self._get("qpus", headers={"x-access-token": token.get()})
        data = response.json()
        all_qpus = {}
        for qpu in data:
            name = qpu["name"]
            del qpu["name"]
            all_qpus[name] = qpu
        return all_qpus

    def list_qpus(self) -> list[str]:
        """
        Returns a list of all accessible QPUs.
        """
        return list(self.get_all_qpus().keys())

    def get_qpu_details(self, name: str) -> ArtemisQPU:
        """
        Returns all details for a particular QPU>
        """
        all_qpus = self.get_all_qpus()
        if name not in all_qpus:
            raise ValueError("QPU with provided name not found.")
        id_ = all_qpus[name]["id"]
        response = self._get(
            f"Qpus/{id_}", headers={"x-access-token": token.get()}
        )
        details = response.json()
        details["maxSamples"] = {
            items["item1"]: items["item2"] for items in details["maxSamples"]
        }
        return ArtemisQPU(
            **{ARTEMIS_QPU_DETAILS_MAP[k]: v for k, v in details.items()}
        )

    def _test_connection(self) -> None:
        """Checks that a connection can be established with the API."""
        self._get_no_catch("jobs")

    def _get_certificate(self) -> Path:
        """
        Used for recovering the certificate required to communicate with the
        remote client. If not found it instructs the user where this should be
        placed.
        """
        # Get persistent data path
        path = get_persistent_file_data_location()
        # Then try to find certificate
        full_path = path / "certificate.pem"
        if not full_path.exists():
            msg = (
                "certificate.pem file not found, this can be added with the "
                "install_certificate method or manually placed in "
                f"the following directory: '{path}'."
            )
            raise FileNotFoundError(msg)
        return full_path



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from collections import UserDict
from types import NoneType
from typing import Any

from lightworks_remote.http import Status
from lightworks_remote.results import QPUSamplingResult
from lightworks_remote.utils import (
    JobError,
    ResultsError,
    get_data_from_json,
    saved_job_exists,
    write_data_to_json,
)

from .job import Job


class BatchJob(UserDict[str, Job]):
    """
    Used for managing a batch of submitted jobs. It allows the status and
    results of all jobs contained to retrieved simultaneously.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__results_cache: dict[str, QPUSamplingResult | str | None] = {}

    @property
    def names(self) -> list[str]:
        """Lists all names of jobs in the batch."""
        return list(self.keys())

    @property
    def jobs(self) -> list[Job]:
        """Returns a list of the created jobs."""
        return list(self.values())

    @property
    def job_id(self) -> dict[str, int]:
        """Returns a dictionary of job names and job ids."""
        return {name: job.job_id for name, job in self.items()}

    @property
    def status(self) -> dict[str, Status]:
        """Returns a dictionary of job names and job statuses."""
        return {name: job.status for name, job in self.items()}

    @property
    def queue_position(self) -> dict[str, int | None]:
        """Returns a dictionary of job names and queue positions."""
        return {name: job.queue_position for name, job in self.items()}

    @property
    def all_complete(self) -> bool:
        """Checks if all jobs are complete."""
        return all(job.complete for job in self.values())

    @property
    def all_success(self) -> bool:
        """Checks all jobs completed successfully."""
        return all(job.success for job in self.values())

    def __str__(self) -> str:
        return f"BatchJob(names = {self.names})"

    def __repr__(self) -> str:
        return "lightworks_remote." + str(self)

    def __getitem__(self, key: str) -> Any:
        if key not in self:
            raise KeyError("Job with matching name not found.")
        return super().__getitem__(key)

    def __setitem__(self, key: str, value: Job) -> None:
        if key in self:
            raise KeyError("Job with provided name already exists.")
        if not isinstance(value, Job):
            raise TypeError("job should be a Job object.")
        if not isinstance(key, str):
            raise TypeError("name should be a string.")
        super().__setitem__(key, value)

    def _add(self, job: Job, name: str) -> None:
        self[name] = job

    def _submit_all(self) -> None:
        """Submits all of the jobs contained within the batch."""
        for name, job in self.items():
            job._submit()
            self.__results_cache[name] = None

    def cancel_all(self) -> None:
        """
        Cancels all submitted jobs within a batch.
        """
        for job in reversed(list(self.values())):
            if job.complete:
                continue  # Skip any jobs already completed
            try:
                job.cancel()
            except JobError as e:
                print(str(e))  # noqa: T201

    def wait_until_complete(self) -> None:
        """Pauses Python execution until all jobs are completed."""
        # Work through each job until all are finished.
        for job in self.values():
            job.wait_until_complete()

    def get_all_results(self) -> dict[str, QPUSamplingResult | str | None]:
        """
        Retrieve all results from jobs.
        """
        if not self.all_complete:
            raise ResultsError("Jobs are not all complete yet.")
        for name in self.names:
            self.get_result(name)

        return self.__results_cache

    def get_result(self, name: str) -> QPUSamplingResult | str:
        """
        Retrieve a singular result using the job name as a reference.
        """
        if name not in self.__results_cache or isinstance(
            self.__results_cache[name], str | NoneType
        ):
            try:
                self.__results_cache[name] = self[name].get_result()
            # If an error is raised then use exception string as result and log
            # an error as well
            except ResultsError as e:
                self.__results_cache[name] = str(e)
                warnings.warn(
                    f"Job '{name}' (id = {self[name].job_id}) did not generate "
                    "any results.",
                    stacklevel=2,
                )

        return self.__results_cache[name]  # type: ignore[return-value]

    def retry_failed(self) -> bool:
        """
        Resubmits any jobs which may have failed.

        Returns:

            bool : Indicates whether any jobs were re-submitted or if all were
                successful.

        """
        job_failed = False
        for name, job in self.items():
            if not job.success:
                job_failed = True
                job.retry()
                self.__results_cache[name] = None
        return job_failed

    def save_to_file(
        self, filename: str, allow_overwrite: bool = False
    ) -> None:
        """
        Saves the job id and payload to a file so that this can be restored at a
        later date.

        Args:

            filename (str) : The name to use for saving the job data, this can
                also contain a path to save it within a specific directory.

            allow_overwrite (bool) : Controls whether an exists saved job while
                be overwritten. Defaults to False.

        """
        if saved_job_exists(filename) and not allow_overwrite:
            raise ValueError(
                "A saved job with this name already exists. Either use a "
                "different name or set allow_overwrite to True to allow "
                "replacement."
            )
        data: dict[str, Any] = {
            name: job._get_job_data() for name, job in self.items()
        }
        data["job_type"] = "batch_job"

        write_data_to_json(data, filename)

    def load_from_file(self, filename: str) -> None:
        """
        Loads job data from the provided file path.

        Args:

            filename (str) : The name of the data file to load, this can also
                contain a path to load from within a specific directory.

        """
        data = get_data_from_json(filename)

        super().__init__()  # Clear jobs before saving
        try:
            for name, job_data in data.items():
                if name == "job_type":
                    continue
                self[name] = Job()
                self[name]._job_id = job_data["job_id"]
                self[name]._payload = job_data["payload"]
                self.__results_cache[name] = None
        except KeyError as e:
            raise JobError(
                "Unable to locate required job data within file."
            ) from e



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from collections.abc import Callable
from copy import copy
from types import FunctionType, MethodType
from typing import Any

import matplotlib.figure
import matplotlib.pyplot as plt
from lightworks import State
from lightworks.sdk.results import Result
from lightworks.sdk.utils import PostSelectionFunction
from lightworks.sdk.utils.heralding import remove_heralds_from_state
from lightworks.sdk.utils.post_selection import PostSelectionType

from lightworks_remote.utils import write_data_to_json


class QPUSamplingResult(Result[State, int]):
    """
    Stores the results from a job, handling the application of post-selection
    rules if this is included as a "post_select" or "herald" field in the data
    dictionary.

    Args:

        data (dict) : The data to process and store in the Result object.

    """

    def __init__(self, data: dict[str, Any]) -> None:
        if data["results"] is None:
            super().__init__({})
            self.__outputs = None
            self.__raw_data = data
            self._post_selection = data.get("job_data")
            warnings.warn(
                "No results were returned by the system.", stacklevel=1
            )
            return
        # Copy data to avoid modification issues
        data = copy(data)
        # Extract post-selection from data if included
        job_data = data.get("job_data")
        # Apply post_selection to get data processed data
        if job_data is not None:
            post_select = job_data.pop("post_select", [])
            herald = job_data.pop("herald", [])
            post_select = [] if post_select is None else post_select
            herald = [] if herald is None else herald
            herald_modes = [h[0][0] for h in herald]
            rules = list(post_select) + list(herald)
            try:
                data_r = apply_post_selection(
                    data["results"], rules, herald_modes
                )
            except IndexError:
                warnings.warn(
                    "Raw data returned: An index error was raised while trying "
                    "to apply post-selection and/or heralding to the results. "
                    "This likely results from one of the criteria being "
                    "outside of the allowed range.",
                    stacklevel=2,
                )
                data_r = data["results"]
            except Exception as e:  # noqa: BLE001
                warnings.warn(
                    "Raw data returned: An unknown error occurred while trying "
                    "to apply post-selection and/or heralding to the results. "
                    f"Returning raw data instead. Exception: {e}.",
                    stacklevel=2,
                )
                data_r = data["results"]
        else:
            data_r = data["results"]
        super().__init__(data_r)
        # Also save raw data and post-selection
        self.__raw_data = data["results"]
        self._post_selection = (
            {"post_select": post_select, "herald": herald}
            if job_data is not None
            else None
        )
        self._job_data = job_data
        # And calculate outputs from data
        self.__outputs = list(self.keys())

    def __getitem__(self, item: State) -> int:
        """Custom get item behaviour - used when object accessed with []."""
        if not self:
            raise ValueError("Result contains no data.")
        if not isinstance(item, State):
            raise TypeError("Get item value must be a State.")
        if item not in self:
            raise KeyError("Provided output state not in data.")
        return super().__getitem__(item)

    @property
    def raw_data(self) -> dict[str, Any]:
        """
        Stores raw data from a job.
        """
        return self.__raw_data

    @property
    def outputs(self) -> list[State] | None:
        """
        All outputs measured in the result.
        """
        return self.__outputs

    @property
    def post_selection(
        self,
    ) -> list[tuple[tuple[int, ...], tuple[int, ...]]] | None:
        """
        Returns any post-selection criteria included with a result.
        """
        return self._post_selection

    @property
    def job_data(self) -> dict[str, Any] | None:
        """Returns any additional data included within a job."""
        return self._job_data

    def plot(
        self, show: bool = True
    ) -> tuple[matplotlib.figure.Figure, plt.Axes] | None:
        """
        Creates a bar chart plot of the data contained within the result.
        """
        if not len(self):
            raise ValueError("Result contains no data.")
        fig, ax = plt.subplots(figsize=(7, 6))
        x_data = range(len(self))
        ax.bar(x_data, list(self.values()))
        ax.set_xticks(x_data)
        labels = [str(s) for s in self]
        ax.set_xticklabels(labels, rotation=90)
        ax.set_xlabel("State")
        ax.set_ylabel("Counts")

        # Optionally use show on plot if specified
        if show:
            plt.show()
            return None
        return (fig, ax)

    def save(self, filename: str, raw_data: bool = False) -> None:
        """
        Saves the stored data to a json file.

        Args:

            filename (str) : The name to use for saving the results data,
                this can also contain a path to save it within a specific
                directory.

            raw_data (bool, optional) : Controls whether the post-processed or
                raw data should be saved.

        """
        # Select correct data and save
        if not raw_data:
            if not len(self):
                raise ValueError("Result contains no data.")
            data = {str(k.s)[1:-1]: v for k, v in self.items()}
        else:
            data = self.raw_data

        write_data_to_json(data, filename)

    def map(
        self,
        mapping: Callable[[State, Any], State],
        *args: Any,
        **kwargs: Any,
    ) -> "QPUSamplingResult":
        """
        Performs a generic remapping of states based on a provided function.
        """
        if not isinstance(mapping, FunctionType | MethodType):
            raise TypeError(
                "Provided mapping should be a callable function which accepts "
                "and returns a State object."
            )
        mapped_result: dict[State, int] = {}
        for out_state, val in self.items():
            new_s = mapping(out_state, *args, **kwargs)
            if new_s in mapped_result:
                mapped_result[new_s] += val
            else:
                mapped_result[new_s] = val
        return self._recombine_mapped_result(mapped_result)

    def apply_post_selection(
        self, post_selection: PostSelectionType | Callable[[State], bool]
    ) -> "QPUSamplingResult":
        """
        Applies an additional post-selection criteria to the stored result and
        returns this as a new object.
        """
        if isinstance(post_selection, FunctionType | MethodType):
            post_selection = PostSelectionFunction(post_selection)
        if not isinstance(post_selection, PostSelectionType):
            raise TypeError(
                "Provided post_selection should either be a PostSelection "
                "object or a callable function which accepts a state and "
                "returns a boolean to indicate whether the state is valid."
            )
        return self._recombine_mapped_result(
            {
                state: val
                for state, val in self.items()
                if post_selection.validate(state)
            }
        )

    def _recombine_mapped_result(
        self, mapped_result: dict[State, int]
    ) -> "QPUSamplingResult":
        """Creates a new Result object from mapped data."""
        return QPUSamplingResult({"results": mapped_result})


def apply_post_selection(
    data: dict[State, int],
    rules: list[tuple[tuple[int, ...], tuple[int, ...]]],
    herald_modes: list[int],
) -> dict[State, int]:
    """
    Applies the post-selection criteria to a set of data in a dictionary.

    Args:

        data (dict) : The unprocessed data dictionary.

        rules (list) : The post-selection rules to be applied.

        herald_modes (list) : The modes used for heralding, which will be
            removed from the state.

    Returns:

        dict : The processed dictionary of data.

    """
    return {
        State(remove_heralds_from_state(k, herald_modes)): v
        for k, v in data.items()
        if meets_post_selection(k, rules)
    }


def meets_post_selection(
    state: State, post_selection: list[tuple[tuple[int, ...], tuple[int, ...]]]
) -> bool:
    """
    Validates whether a provided State meets the set post-selection criteria.

    Args:

        state (State) : The state which is to be checked.

        post_selection (int) : The post-selection rules to be applied.

    Returns:

        bool : Indicates whether the provided State meets the post-selection
            criteria.

    """
    for rule in post_selection:
        total = 0
        for mode in rule[0]:
            total += state[mode]
        if total not in rule[1]:
            return False
    return True



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .batch_job import BatchJob
from .job import Job
from .job_management import (
    cancel_job,
    check_job_complete,
    get_job_details,
    get_job_status,
    get_results,
    list_all_jobs,
    list_scheduled_jobs,
    load_job_from_file,
    show_scheduled,
)
from .qpu import QPU
from .system_management import get_qpu_details, list_qpus

__all__ = [
    "QPU",
    "BatchJob",
    "Job",
    "cancel_job",
    "check_job_complete",
    "get_job_details",
    "get_job_status",
    "get_qpu_details",
    "get_results",
    "list_all_jobs",
    "list_qpus",
    "list_scheduled_jobs",
    "load_job_from_file",
    "show_scheduled",
]



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

import numpy as np
import pytest
from lightworks import (
    PhotonicCircuit,
    Sampler,
    State,
    random_unitary,
)

from lightworks_remote.payloads import Payload
from lightworks_remote.remote.sampler_compiler import SamplerCompiler
from lightworks_remote.utils.conversion import convert_fock_data_to_state
from lightworks_remote.utils.file_management import (
    get_data_from_json,
    validate_json_filename,
    write_data_to_json,
)
from lightworks_remote.utils.serialization import serialize


class TestSerialize:
    """
    Unit tests for the serialization functionality of the remote module.
    """

    def test_payload(self):
        """
        Checks that a generated payload can be serialized.
        """
        task = Sampler(PhotonicCircuit(4), State([1, 0, 1, 0]), 10000)
        compiler = SamplerCompiler(task._generate_task())
        payload = Payload(**compiler._generate_payload())

        serialize(payload)

    def test_payload_to_json(self):
        """
        Checks that a generated payload can be serialized and then saved to a
        json file.
        """
        task = Sampler(PhotonicCircuit(4), State([1, 0, 1, 0]), 10000)
        compiler = SamplerCompiler(task._generate_task())
        payload = Payload(**compiler._generate_payload())
        payload = serialize(payload)
        with open("test_payload_dump.json", "w") as f:  # noqa: PLW1514
            json.dump(payload, f)
        os.remove("test_payload_dump.json")

    @pytest.mark.parametrize("value", [1, 1.1, "Test", True, [1, 2.4]])
    def test_serialize_unmodified(self, value):
        """
        Checks serialization does nothing for certain types.
        """
        assert serialize(value) == value

    def test_serialize_complex(self):
        """
        Checks serialization of complex number converts it to a string.
        """
        number = 1.9 + 0.5j
        converted = serialize(number)
        assert isinstance(converted, str)
        assert str(number.real) in converted
        assert str(number.imag) in converted

    def test_serialize_state(self):
        """
        Checks that a provided state is converted to a list by the serialize
        method.
        """
        state = serialize(State([1, 0, 1, 0]))
        assert isinstance(state, list)

    def test_serialize_list(self):
        """
        Checks serialization on a list of values will serialize all of the
        values.
        """
        converted = serialize([1.1, State([1, 2, 3, 4]), 1 + 0.5j])
        assert converted == [1.1, [1, 2, 3, 4], serialize(1 + 0.5j)]

    def test_serialize_array(self):
        """
        Checks serialization converts a numpy array into a list.
        """
        array = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        converted = serialize(array)
        assert isinstance(converted, list)
        assert converted == array.tolist()

    def test_serialize_complex_array(self):
        """
        Checks serialization of complex array converts it into a list with
        string values.
        """
        array = random_unitary(3)
        converted = serialize(array)
        assert isinstance(converted, list)
        assert isinstance(converted[0][0], str)
        assert converted == array.astype(str).tolist()

    def test_serialize_dict(self):
        """
        Checks serialization of dictionary operates on the keys/values.
        """
        d = {State([1, 0, 1, 0]): 1 + 0.5j, "test": random_unitary(3)}
        converted = serialize(d)
        assert "[1, 0, 1, 0]" in converted
        assert converted["[1, 0, 1, 0]"] == serialize(1 + 0.5j)
        assert "test" in converted
        assert isinstance(converted["test"], list)

    def test_serialize_dict_nested(self):
        """
        Checks serialization of dictionary operates on the keys/values when
        a dictionary is one of the values..
        """
        d = {
            State([1, 0, 1, 0]): 1 + 0.5j,
            "test": {State([1, 0, 1, 0]): 1 + 0.5j, "test": random_unitary(3)},
        }
        converted = serialize(d)
        assert "[1, 0, 1, 0]" in converted
        assert converted["[1, 0, 1, 0]"] == serialize(1 + 0.5j)
        assert "[1, 0, 1, 0]" in converted["test"]
        assert isinstance(converted["test"]["test"], list)


class TestFileManagement:
    """
    Unit tests for file_management.py section of utils.
    """

    @pytest.mark.parametrize("value", [["test"], True, 1])
    def test_validate_file_name(self, value):
        """
        Checks that validate file name raises an exception if a non-string value
        is provided.
        """
        with pytest.raises(TypeError):
            validate_json_filename(value)

    def test_validate_file_name_adds_json(self):
        """
        Checks that validate file name adds the json extension if this is not
        included.
        """
        filename = str(validate_json_filename("test"))
        assert filename.endswith(".json")

    def test_validate_file_name_json_included(self):
        """
        Checks that validate file name doesn't add the json extension if this is
        already included.
        """
        filename = str(validate_json_filename("test.json"))
        assert filename.endswith(".json")
        assert not filename.endswith(".json.json")

    def test_write_data(self):
        """
        Checks that write data is able to write a dictionary to a json file.
        """
        data = {"test": [1, 2, 3], "test2": [4, 5, 6]}
        write_data_to_json(data, "_test_data.json")
        assert os.path.isfile("_test_data.json")
        os.remove("_test_data.json")

    def test_write_data_no_json(self):
        """
        Checks that write data is able to write a dictionary to a json file when
        the extension is not provided.
        """
        data = {"test": [1, 2, 3], "test2": [4, 5, 6]}
        write_data_to_json(data, "_test_data")
        assert os.path.isfile("_test_data.json")
        os.remove("_test_data.json")

    def test_write_data_path(self):
        """
        Checks that write data is able to write a dictionary to a json file
        when a path is included.
        """
        data = {"test": [1, 2, 3], "test2": [4, 5, 6]}
        write_data_to_json(data, ".pytest_cache/_test_data.json")
        assert os.path.isfile(".pytest_cache/_test_data.json")
        os.remove(".pytest_cache/_test_data.json")

    def test_write_data_non_serialized(self):
        """
        Checks that write data is able to serialize a dictionary and save to a
        json file.
        """
        data = {"test": State([1, 2, 3]), "array": random_unitary(6)}
        write_data_to_json(data, "_test_data.json")
        assert os.path.isfile("_test_data.json")
        os.remove("_test_data.json")

    def test_get_data(self):
        """
        Checks that get data is able to recover a dictionary from a created json
        file.
        """
        data = {"test3": [1, 2, 3], "test4": [4, 5, 6]}
        with open("_test_data.json", "w", encoding="locale") as f:
            json.dump(data, f, indent=4)
        recovered = get_data_from_json("_test_data.json")
        assert recovered == data
        os.remove("_test_data.json")

    def test_get_data_no_json(self):
        """
        Checks that get data is able to recover a dictionary from a created json
        file when the file extension is not provided
        """
        data = {"test3": [1, 2, 3], "test4": [4, 5, 6]}
        with open("_test_data.json", "w", encoding="locale") as f:
            json.dump(data, f, indent=4)
        recovered = get_data_from_json("_test_data")
        assert recovered == data
        os.remove("_test_data.json")

    def test_get_data_path(self):
        """
        Checks that get data is able to recover a dictionary from a created json
        file within a sub-directory.
        """
        data = {"test3": [1, 2, 3], "test4": [4, 5, 6]}
        with open(".pytest_cache/_test_data.json", "w", encoding="locale") as f:
            json.dump(data, f, indent=4)
        recovered = get_data_from_json(".pytest_cache/_test_data.json")
        assert recovered == data
        os.remove(".pytest_cache/_test_data.json")


class TestUtils:
    """
    For the testing of general utilities.
    """

    @pytest.mark.parametrize(
        ("input_data", "output_state"),
        [
            ("1,", State([1])),
            ("1, ", State([1])),
            ("(1,)", State([1])),
            ("(1, )", State([1])),
            ("1, 2, 3", State([1, 2, 3])),
            ("(1, 2, 3)", State([1, 2, 3])),
        ],
    )
    def test_state_conversion(self, input_data, output_state):
        """
        Checks that the convert_fock_data_to_state correctly processes various
        data formats to the correct Lightworks state.
        """
        assert convert_fock_data_to_state(input_data) == output_state
